#include <linux/config.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/smp.h>

static int main_processor;
static int slave_tested;

static volatile int our_async_method_started = 0;
static volatile int our_async_method_finished = 0;
static volatile int our_sync_method_started = 0;
static volatile int our_sync_method_finished = 0;
static volatile int async_start_state_when_sync_triggered = 0;
static volatile int async_finish_state_when_sync_triggered = 0;

static void our_async_method(void* ptr) {
	/* Not us... */
	if (smp_processor_id() != slave_tested) {
		return;
	}
	/* Mark we started... */
	our_async_method_started = 1;
	/* Wait... */
	udelay(200);
	/* Mark we are done... */
	our_async_method_finished = 1;
}

static void our_sync_method(void* ptr) {
	/* Not us... */
	if (smp_processor_id() != slave_tested) {
		return;
	}
	async_start_state_when_sync_triggered = our_async_method_started;
	async_finish_state_when_sync_triggered = our_async_method_finished;
	/* Mark we started... */
	our_sync_method_started = 1;
	/* Wait... */
	udelay(200);
	/* Mark we are done... */
	our_sync_method_finished = 1;
}

static int loadme(void) {
	unsigned long jifstart;
	int async_was_async;
	int sync_was_async;
	
	main_processor = smp_processor_id();
	
	/* If we are processor 0, run our IPI tests on processor 1,
	   otherwise run them on processor 0. */
	if (main_processor == 0) {
		slave_tested = 1;
	} else {
		slave_tested = 0;
	}
	if (!cpu_online(slave_tested)) {
		printk(KERN_ERR "Sorry, processor %u is not online.\n", slave_tested);
		return -EBUSY;
	}
	if (!cpu_online(main_processor)) {
		printk(KERN_ERR "Sorry, processor %u is not online.\n", main_processor);
		return -EBUSY;
	}
	/* Wait for jiffies increment - we'll hope that no other 
	   interrupts will occur quickly after timer interrupt */
	jifstart = jiffies;
	while (jifstart == jiffies) ;
	smp_call_function(our_async_method, NULL, 0, 0);
	async_was_async = (our_async_method_finished == 0);
	smp_call_function(our_sync_method, NULL, 1, 1);
	sync_was_async = (our_sync_method_finished == 0);
	
	printk(KERN_ERR "Test done. ASYNC was %s, SYNC was %s, nesting %s.\n",
			async_was_async ? "ASYNC" : "SYNC",
			sync_was_async ? "ASYNC" : "SYNC",
			async_finish_state_when_sync_triggered ? "did not happen" : "occured");
	if (async_was_async == 1 &&
	    sync_was_async == 0 &&
	    async_finish_state_when_sync_triggered == 1) {
		printk(KERN_ERR "Test passed.\n");
	} else {
		printk(KERN_ERR "Test failed.\n");
	}
	/* Wait for IPIs to finish... */
	if (our_async_method_started) {
		while (!our_async_method_finished) {
			schedule();
		}
	} else {
		printk(KERN_ERR "ASYNC IPI was completely lost!\n");
	}
	if (our_sync_method_started) {
		while (!our_sync_method_finished) {
			schedule();
		}
	}
	return -EBUSY;
}

module_init(loadme);
 
